/*-
 * Copyright (C)2017..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2017..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: func.c,v 1.21 2018/07/29 22:33:42 babolo Exp $"

#define BLIN_COMPAT      3
#define Bpars_COMPAT     3
#define MIFE_COMPAT      4
#define RECOBE_COMPAT    4
#define PGOBLIN_COMPAT   2
#define PGOBLIN_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

const u_char *
/************************************************************************
 **                                                                    **/
winpoint(mife_descriptor *md, off_t offset, ssize_t trd) {            /**
 **                                                                    **
 ************************************************************************/
    const u_char    *p = NULL;
    ssize_t          rdr;

    errno = 0;
    if  (trd > (rdr = mife_read(md, trd, offset))) {
        if  (!!mife_ctlsize(md)) {
            ifBLIN_QW0("mife_read %"BLIN_D" < %"BLIN_D, rdr, trd);
            if  (rdr >= 0) errno = EFTYPE;
        }
    } else if (!(p = mife_pointer(md))) {
        ifBLIN_QW0("mife_pointer");
    }
    return(p);
}

int
/************************************************************************
 ************************************************************************
 **                                                                    **/
pgoblin_breg(pgoblin_command cmd, pgoblin_nr rn) {                    /**
 **                                                                    **
 ************************************************************************
 **      cmd   rn.     **
 **    cmd  rn  ,   < 0 **
 ************************************************************************
 ************************************************************************/
    int j;

    j = -1;
    if  (pgoblin.a[cmd].flags & pgoblin.p[rn].sym) {
        /*  a[cmd].flags    [rn] */
        j = (pgoblin.p[rn].bin < 0)
            /*      */
          ? (pgoblin.a[cmd].flags >> ~pgoblin.p[rn].bin) & pgoblin.p[rn].mask
            /*      */
          : pgoblin.p[rn].bin
        ;
        if  (!(pgoblin.a[cmd].ch[j] & pgoblin.p[rn].val) && !!pgoblin.a[cmd].ch[j]) {
            /*                            *
             *     ,  -     . *
             *    0 j ,   j  ,      *
             *        j                    *
             */
            j = -2;
    }   }
    return(j);
}

void
/**********************************************************************
 **                                                                  **/
putnum(pgoblin_main *options, u_int32_t num) {                      /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    u_char  buff[10];
    ssize_t wrt;
    int     l;

    l = recobe_cstore(buff, num);
    errno = 0;
    if  (l != (wrt = mife_writ(options->xtrace, buff, l))) {
        ifBLIN_QW0("mife_writ %d <> %d", wrt, l);
}   }
#   undef blin_internal_flags

void
/**********************************************************************
 **                                                                  **/
putext(pgoblin_main *options, char *text, size_t length, int bin) { /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    char    null = 0xFF;
    ssize_t strl;
    ssize_t wrt;

    if  (!text) {
        errno = 0;
        if  (1 != (wrt = mife_writ(options->xtrace, &null, 1))) {
            ifBLIN_QW0("mife_writ %d <> 1", wrt);
        }
    } else {
        if (!!bin || !!length) {
           strl = length;
        } else {
           strl = strlen(text);
        }
        if  (strl > 0x000000007FFFFFFFLL) {
            ifBLIN_QX0("string too big %"BLIN_X, strl);
        } else {
            putnum(options, strl);
            errno = 0;
            if  (strl != (wrt = mife_writ(options->xtrace, text, strl))) {
                ifBLIN_QW0("mife_writ %"BLIN_D" <> %"BLIN_D, wrt, strl);
}   }   }   }
#   undef blin_internal_flags

int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_cmdexe(int cmd, pgoblin_exenv *exenv, pgoblin_nr *rn) {     /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    u_char         buf[PGOBLIN_GOEXECU];
    char           null = 0xFF;
    u_int8_t       binreg;
    u_int8_t       symreg;
    ssize_t        wrt;
    int            ex;

    ifBLIN_QX3( "+%s %c%c%c%c%c%c %08X"
              , pgoblin.a[cmd].str
              , pgoblin_regn[rn[0]]
              , pgoblin_regn[rn[1]]
              , pgoblin_regn[rn[2]]
              , pgoblin_regn[rn[3]]
              , pgoblin_regn[rn[4]]
              , pgoblin_regn[rn[5]]
              , pgoblin.a[cmd].flags
              );
    if  (0 <= exenv->options->xtrace) {
        {   pgoblin_exenv *c;
            int32_t        d;
            int32_t        s;

            for (s = 0, c = exenv; c; c = c->stat) s++;
            for (d = 0, c = exenv; c; c = c->dyna) d++;
            bzero(PGOBLIN_GOSHORT + buf, pgoblin.binparam);
            buf[0] = (cmd >> 8) | 0x0E0;
            buf[1] = cmd & 0x0FF;
            recobe_store(&buf[PGOBLIN_HEDLN], PGOBLIN_PIDLN, getpid());
            for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
                if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
                    buf[PGOBLIN_GOSHORT + binreg] = rn[symreg];
            }   }
            recobe_store(&buf[PGOBLIN_GOCMD], sizeof(int32_t), s);
            recobe_store(&buf[sizeof(int32_t) + PGOBLIN_GOCMD], sizeof(int32_t), d);
            recobe_store(&buf[sizeof(int32_t) * 2 + PGOBLIN_GOCMD], sizeof(int32_t), exenv->ppoint);
        }
        if  (!!flock(exenv->options->xtrace, LOCK_EX)) {
            ifBLIN_QW0("xtrace lock");
        } else {
            errno = 0;
            if  (PGOBLIN_GOEXECU != (wrt = mife_writ(exenv->options->xtrace, buf, PGOBLIN_GOEXECU))) {
                ifBLIN_QW0("mife_writ %d <> %d", wrt, PGOBLIN_GOEXECU);
            }
            for (binreg = 0; binreg < pgoblin.binparam; binreg++) {
                pgoblin_io *io = &exenv->options->io[buf[PGOBLIN_GOSHORT + binreg] & 0x03F];

                if  (PGOBLIN_PATHPARM & pgoblin.a[cmd].ch[binreg]) {
                    putext(exenv->options, io->path, 0, 0);
                }
                if  (PGOBLIN_TEXTPARM & pgoblin.a[cmd].ch[binreg]) {
                    putext(exenv->options, io->text, io->length, io->flags & PGOBLIN_BINPARM);
                }
                if  (PGOBLIN_MIFEDES & pgoblin.a[cmd].ch[binreg]) {
                }
                if  (PGOBLIN_PQRESULT & pgoblin.a[cmd].ch[binreg]) {
                    int     row;
                    int     col;
                    int32_t v;

                    if  (!io->pq) {
                        errno = 0;
                        if  (1 != (wrt = mife_writ(exenv->options->xtrace, &null, 1))) {
                            ifBLIN_QW0("mife_writ %d <> 1", wrt);
                        }
                    } else {
                        if  (0 > (v = pgoblin_db_resinfo(PGOBLIN_Ntuples, io->pq))) {
                            ifBLIN_QW0("resinfo Ntuples");
                        } else {
                            putnum(exenv->options, v);
                        }
                        if  (0 > (v = pgoblin_db_resinfo(PGOBLIN_Nfields, io->pq))) {
                            ifBLIN_QW0("resinfo Nfields");
                        } else {
                            putnum(exenv->options, v);
                        }
                        for (row = 0; row < pgoblin_db_resinfo(PGOBLIN_Ntuples, io->pq); ++row) {
                            for (col = 0; col < v; ++col) {
                                if  (pgoblin_db_valinfo(PGOBLIN_IsNull, io->pq, row, col)) {
                                    putext(exenv->options, NULL, 0, 0);
                                } else {
                                    putext( exenv->options
                                          , pgoblin_db_getvalue(io->pq, row, col)
                                          , pgoblin_db_valinfo(PGOBLIN_Length, io->pq, row, col)
                                          , 0
                                          );
                }   }   }   }   }
                if  (PGOBLIN_CONNECT & pgoblin.a[cmd].ch[binreg]) {
                }
                if  (PGOBLIN_JOB & pgoblin.a[cmd].ch[binreg]) {
                }
                if  (PGOBLIN_STYLE & pgoblin.a[cmd].ch[binreg]) {
        }   }   }
        if  (!!flock(exenv->options->xtrace, LOCK_UN)) {
            ifBLIN_QW0("xtrace unlock");
    }   }
    ex = (*(pgoblin.a[cmd].car))(exenv, rn);
    ifBLIN_QX3( "- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

void *
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_getr(pgoblin_main *options, pgoblin_nr rn, int reg) {       /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    switch(reg) {
    case PGO_CSTY: return(&(options->style[rn]));
    case PGO_CJOB: return(&(options->job[rn]));
    case PGO_CCON: return(&(options->conn[rn]));
    default:       return(&(options->io[rn]));
#   undef blin_internal_flags
}   }

void
/*****************************************************************************
 *****************************************************************************
 **                                                                         **/
pgoblin_go( void *x           /*  (f > 0)   (f < 0)     */
/**/      , int f             /*    pgoblin_syntax.ch[]  *-1 */
/**/      , const char *file  /*                            */
/**/      , const char *func  /*                                   */
/**/      , int line          /*                        */
/**/      ) {                                                              /**
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/
#   define blin_internal_flags (pgoblin_options->flags & BLIN_MASK)
    if  (0 <= pgoblin_options->xtrace) {
        u_char    buf[PGOBLIN_GOSHORT];
        u_int64_t d;
        u_int64_t p;
        u_int64_t q;
        int       i;

        if  (0 < f) {
            switch(f) {
            case PGOBLIN_STYLE   :
            case PGOBLIN_JOB     :
            case PGOBLIN_CONNECT :
            case PGOBLIN_OUTSET  :
            case PGOBLIN_PQRESULT:
            case PGOBLIN_MIFEDES :
            case PGOBLIN_TEXTPARM:
            case PGOBLIN_PATHPARM:
                buf[0] = f;
                buf[1] = (u_int64_t)x;
                break;
            default              :
                buf[0] = 0;
                buf[1] = f;
            }
        } else {
            switch(-f) {
            case PGOBLIN_STYLE   :
                p = (u_int64_t)&pgoblin_options->style[0];
                q = (u_int64_t)&pgoblin_options->style[1];
                break;
            case PGOBLIN_JOB     :
                p = (u_int64_t)&pgoblin_options->job[0];
                q = (u_int64_t)&pgoblin_options->job[1];
                break;
            case PGOBLIN_CONNECT :
                p = (u_int64_t)&pgoblin_options->conn[0];
                q = (u_int64_t)&pgoblin_options->conn[1];
                break;
            case PGOBLIN_OUTSET  :
            case PGOBLIN_PQRESULT:
            case PGOBLIN_MIFEDES :
            case PGOBLIN_TEXTPARM:
            case PGOBLIN_PATHPARM:
                p = (u_int64_t)&pgoblin_options->io[0];
                q = (u_int64_t)&pgoblin_options->io[1];
                break;
            default              :
                buf[0] = 0;
                buf[1] = -f;
                goto out;
            }
            d = ((u_int64_t)x - p) / (q - p);
            if  (((d * (q - p)) != ((u_int64_t)x - p)) || !!(d & ~0x3FUL)) {
                ifBLIN_QX1( "@%s:%s:%d: "
                            "Lost syn %"BLIN_X" i%"BLIN_X" c%"BLIN_X" j%"BLIN_X" s%"BLIN_X" f%02X"
                          , file
                          , func
                          , line
                          , BLIN_I(x)
                          , BLIN_I(&pgoblin_options->io[0])
                          , BLIN_I(&pgoblin_options->conn[0])
                          , BLIN_I(&pgoblin_options->job[0])
                          , BLIN_I(&pgoblin_options->style[0])
                          , -f
                          );
                d = 0x3F;
            }
            buf[0] = -f;
            buf[1] = d;
        }
out:
        recobe_store(&buf[PGOBLIN_HEDLN], PGOBLIN_PIDLN, getpid());
        if  (!!flock(pgoblin_options->xtrace, LOCK_EX)) {
            ifBLIN_QW0("xtrace lock");
        } else {
            errno = 0;
            if  (PGOBLIN_GOSHORT != (i = mife_writ(pgoblin_options->xtrace, buf, PGOBLIN_GOSHORT))) {
                ifBLIN_QW0("mife_writ %d <> %d", i, PGOBLIN_GOSHORT);
            }
            if  (!!flock(pgoblin_options->xtrace, LOCK_UN)) {
                ifBLIN_QW0("xtrace unlock");
    }   }   }
#   undef blin_internal_flags
}

static char P[] = "Xftmqocjs";

static ssize_t
/*****************************************************************************
 **                                                                         **/
cload(pgoblin_main *options, mife_descriptor *md, off_t *offset) {         /**
 **                                                                         **
 *****************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    ssize_t          rdr;
    ssize_t          trd;
    const u_char    *p;

    if  (!(p = winpoint(md, *offset, 1))) {
        ifBLIN_QW0("trace");
        trd = (errno == EFTYPE) ? -EX_DATAERR : -EX_IOERR;
        goto out;
    }
    if  (*p == 0x0FF) {
        ++*offset;
        trd = -1;
        goto out;
    }
    rdr = recobe_clen(*p);
    if  (!(p = winpoint(md, *offset, rdr))) {
        ifBLIN_QW0("trace");
        trd = (errno == EFTYPE) ? -EX_DATAERR : -EX_IOERR;
        goto out;
    }
    trd = recobe_cload(p, &rdr);
    if  (recobe_clen(*p) != rdr) {
        ifBLIN_QX0("recobe_cload <> recobe_clen");
    }
    *offset += rdr;
out:;
    return(trd);
#   undef blin_internal_flags
}

static int
/*********************************************************************************
 **                                                                             **/
txtprt(pgoblin_main *options, mife_descriptor *md, off_t *offset, char end) {  /**
 **                                                                             **
 *********************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    int              ex = EX_OK;
    ssize_t          trd;
    const u_char    *p;

    if  (-1 > (trd = cload(options, md, offset))) {
        ifBLIN_QW0("trace @ 0x%"BLIN_O"X", *offset);
        ex = -trd;
    } else if (0 > trd) {
        if  ((options->flags & PGOBLIN_T_COMAND) && ('|' == end)) printf("|");
    } else {
        if  (!(p = winpoint(md, *offset, trd))) {
            ifBLIN_QW0("trace @ 0x%"BLIN_O"X", *offset);
            ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
            goto out;
        }
        if  (options->flags & PGOBLIN_T_COMAND) printf("=%.*s%c", (int)trd, p, end);
        *offset += trd;
    }
out:;
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************
 *****************************************************************************
 **                                                                         **/
pgoblin_traceat( pgoblin_main *options                                     /**/
/**/           , mife_descriptor *(*popin)(pgoblin_main *options)          /**/
/**/           , const char *Tflag                                         /**/
/**/           ) {                                                         /**
 **                                                                         **
 *****************************************************************************
 *****************************************************************************/
#   define blin_internal_flags (options->flags & BLIN_MASK)
    u_char           t[pgoblin.binparam];
    u_char           s[pgoblin.symparam];
    int              ex = EX_OK;
    int              seqnum;
    off_t            offset;
    int              binreg;
    int              symreg;
    u_int16_t        cmd;
    mife_descriptor *md;
    const u_char    *p;

    if  (!(md = popin(options))) {
        ifBLIN_QW0("popin mife_init");
        ex = EX_IOERR;
        goto out;
    }
    if  (!Tflag) {
        ifBLIN_QX0("no trace task");
        ex = EX_USAGE;
        errno = EINVAL;
        goto out;
    }
    while(' ' < *Tflag) {
        switch(*(Tflag++)) {
        case 'R': options->flags |= PGOBLIN_T_REGO;    break;
        case 'C': options->flags |= PGOBLIN_T_COMAND;  break;
        case 'E': options->flags |= PGOBLIN_T_EXECUTE; break;
    }   }
    for (offset = 0, seqnum = 0; ; ++seqnum) {
        if  (!(p = winpoint(md, offset, PGOBLIN_GOSHORT))) {
            if  (!mife_ctlsize(md)) break;
            ifBLIN_QW0("trace file");
            ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
            goto out;
        }
        switch(*p) {
        case PGOBLIN_STYLE   : t[0] = P[8]; break;
        case PGOBLIN_JOB     : t[0] = P[7]; break;
        case PGOBLIN_CONNECT : t[0] = P[6]; break;
        case PGOBLIN_OUTSET  : t[0] = P[5]; break;
        case PGOBLIN_PQRESULT: t[0] = P[4]; break;
        case PGOBLIN_MIFEDES : t[0] = P[3]; break;
        case PGOBLIN_TEXTPARM: t[0] = P[2]; break;
        case PGOBLIN_PATHPARM: t[0] = P[1]; break;
        case 0               : t[0] = P[0]; break;
        default              : t[0] = 0;
        }
        if  (!!t[0]) {
            ++p;
            if  (options->flags & PGOBLIN_T_REGO) {
                printf( "R%6d.%u P%c %c\n"
                      , seqnum
                      , (unsigned)recobe_load(p + 1, 3)
                      , t[0]
                      , pgoblin_regn[*p & 0x03F]
                      );
            }
            offset += PGOBLIN_GOSHORT;
            continue;
        }
        switch(*p & 0x0E0) {
        if  ((*p & 0x0E0) != 0x0E0) {
            ifBLIN_QX0("illegal trace file 0x%02X @ 0x%012"BLIN_O"X", *p, offset);
            ex = EX_DATAERR;
            errno = EILSEQ;
            goto out;
        }
        case 0xE0: {
        ;       if  (!(p = winpoint(md, offset, PGOBLIN_GOEXECU))) {
        ;           ifBLIN_QW0("trace file");
        ;           ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;           goto out;
        ;       }
        ;       cmd = (*(p) << 8) & 0x0100;
        ;       cmd |= *(p + 1);
        ;       bzero(t, pgoblin.binparam);
        ;       bzero(s, pgoblin.symparam);
        ;       for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
        ;           if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
        ;               s[symreg] = t[binreg] = p[PGOBLIN_GOSHORT + binreg] & 0x03F;
        ;       }   }
        ;       if  (options->flags & PGOBLIN_T_COMAND) {
        ;           printf( "C%6d.%u (%ds %dd) N%d #%s %c%c%c%c%c%c\n"
                          , seqnum
                          , (unsigned)recobe_load(PGOBLIN_HEDLN + p, PGOBLIN_PIDLN)
                          , (int)recobe_load(PGOBLIN_GOCMD + p, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) + PGOBLIN_GOCMD + p, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) * 2 + PGOBLIN_GOCMD + p, sizeof(int32_t))
                          , pgoblin.a[cmd].str
                          , pgoblin_regn[s[0]]
                          , pgoblin_regn[s[1]]
                          , pgoblin_regn[s[2]]
                          , pgoblin_regn[s[3]]
                          , pgoblin_regn[s[4]]
                          , pgoblin_regn[s[5]]
                          );
        ;       }
        ;       offset += PGOBLIN_GOEXECU;
        ;       for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
        ;           if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
                        u_char ch = pgoblin.a[cmd].ch[binreg];

        ;               if  (PGOBLIN_PATHPARM & ch) {
        ;                   if  (options->flags & PGOBLIN_T_COMAND) {
        ;                       printf( "        %d %d [%c]%sPf"
                                      , binreg
                                      , symreg
                                      , pgoblin_regn[t[binreg] & 0x03F]
                                      , (ch & PGOBLIN_OUTSET) ? "Po|" : "   "
                                      );
        ;                   }
        ;                   if  (!!(ex = txtprt(options, md, &offset, '~'))) {
        ;                       ifBLIN_QW0("trace Pf");
        ;                       goto out;
        ;                   }
        ;                   if  (options->flags & PGOBLIN_T_COMAND) printf("\n");
        ;               }
        ;               if  (PGOBLIN_TEXTPARM & ch) {
        ;                   if  (options->flags & PGOBLIN_T_COMAND) {
        ;                       printf( "        %d %d [%c]%sPt"
                                      , binreg
                                      , symreg
                                      , pgoblin_regn[t[binreg] & 0x03F]
                                      , (ch & PGOBLIN_OUTSET) ? "Po|" : "   "
                                      );
        ;                   }
        ;                   if  (!!(ex = txtprt(options, md, &offset, '~'))) {
        ;                       ifBLIN_QW0("trace Pt");
        ;                       goto out;
        ;                   }
        ;                   if  (options->flags & PGOBLIN_T_COMAND) printf("\n");
        ;               }
        ;               if  (PGOBLIN_MIFEDES  & ch) {
        ;               }
        ;               if  (PGOBLIN_PQRESULT & ch) {
        ;                   if  (options->flags & PGOBLIN_T_COMAND) {
        ;                       printf( "        %d %d [%c]%sPq"
                                      , binreg
                                      , symreg
                                      , pgoblin_regn[t[binreg] & 0x03F]
                                      , (ch & PGOBLIN_OUTSET) ? "Po|" : "   "
                                      );
        ;                   }
        ;                   if  (!(p = winpoint(md, offset, 1))) {
        ;                        ifBLIN_QW0("trace q");
        ;                        ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;                        goto out;
        ;                   }
        ;                   if  (*p == 0x0FF) {
        ;                       if  (options->flags & PGOBLIN_T_COMAND) printf("\n");
        ;                       ++offset;
        ;                   } else {
                                int     row;
                                int     col;

        ;                       row = cload(options, md, &offset); /* 0 > ? XXXX*/
        ;                       col = cload(options, md, &offset); /* 0 > ? XXXX*/
        ;                       if  (options->flags & PGOBLIN_T_COMAND) {
        ;                           printf(" %d rows x %d cols\n", row, col);
        ;                       }
        ;                       for (int r = 0; r < row; ++r) {
        ;                           for (int c = 0; c < col; ++c) {
        ;                               if  (!!(ex = txtprt(options, md, &offset, '|'))) {
        ;                                   ifBLIN_QW0("trace Pq");
        ;                                   goto out;
        ;                           }   }
        ;                           if  (options->flags & PGOBLIN_T_COMAND) printf("\n");
        ;               }   }   }
        ;               if  (PGOBLIN_OUTSET   & ch) {
        ;               }
        ;               if  (PGOBLIN_CONNECT  & ch) {
        ;               }
        ;               if  (PGOBLIN_JOB      & ch) {
        ;               }
        ;               if  (PGOBLIN_STYLE    & ch) {
        ;   }   }   }   }
        ;   break;
        case 0xC0: {
        ;       if  (!(p = winpoint(md, offset, PGOBLIN_GOEXECU))) {
        ;           ifBLIN_QW0("trace file");
        ;           ex = (errno == EFTYPE) ? EX_DATAERR : EX_IOERR;
        ;           goto out;
        ;       }
        ;       cmd = (*(p) << 8) & 0x0100;
        ;       cmd |= *(p + 1);
        ;       bzero(t, pgoblin.binparam);
        ;       bzero(s, pgoblin.symparam);
        ;       for (symreg = 0; symreg < pgoblin.symparam; symreg++) {
        ;           if  (0 <= (binreg = pgoblin_breg(cmd, symreg))) {
        ;               s[symreg] = t[binreg] = p[PGOBLIN_GOSHORT + binreg];
        ;       }   }
        ;       if  (options->flags & PGOBLIN_T_EXECUTE) {
        ;           printf( "E%6d.%u (%ds %dd) N%d #%s %c%c%c%c%c%c\n"
                          , seqnum
                          , (unsigned)recobe_load(PGOBLIN_HEDLN + p, PGOBLIN_PIDLN)
                          , (int)recobe_load(p + PGOBLIN_GOCMD, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) + p + PGOBLIN_GOCMD, sizeof(int32_t))
                          , (int)recobe_load(sizeof(int32_t) * 2 + p + PGOBLIN_GOCMD, sizeof(int32_t))
                          , pgoblin.a[cmd].str
                          , pgoblin_regn[s[0] & 0x03F]
                          , pgoblin_regn[s[1] & 0x03F]
                          , pgoblin_regn[s[2] & 0x03F]
                          , pgoblin_regn[s[3] & 0x03F]
                          , pgoblin_regn[s[4] & 0x03F]
                          , pgoblin_regn[s[5] & 0x03F]
                          )
        ;       }
        ;       offset += PGOBLIN_GOEXECU;
        ;   }
        ;   break;
        default: {
        ;       ifBLIN_QX0("illegal trace file 0x%02X @ 0x%012"BLIN_O"X", *p, offset);
        ;       ex = EX_DATAERR;
        ;       errno = EILSEQ;
        ;       goto out;
    }   }   }
out:;
    return(ex);
#   undef blin_internal_flags
}
